from rest_framework import (
    exceptions,
    status,
    response,
)
from .settings import MB_EXCEPTIONS

ERROR_CODE_NAME = MB_EXCEPTIONS.get(
    'ERROR_CODE_NAME', 'error')
ERROR_MESSAGE_NAME = MB_EXCEPTIONS.get(
    'ERROR_MESSAGE_NAME', 'error_description')


class ErrorMeta(type):
    """
    meta class of errors
    """

    errors = {}
    errors_4_swapping = {}

    def __new__(mcs, class_name, class_bases, class_attr):
        cls = super().__new__(mcs, class_name, class_bases, class_attr)
        if class_name == 'Error':
            return cls

        error_code = getattr(cls, 'error_code', None)
        assert error_code, (
            'error_code is not specified in error class {}'.format(
                class_name)
        )
        assert error_code not in mcs.errors, (
            'duplicated error_code {} found'.format(error_code))

        mcs.errors[error_code] = cls
        swap = getattr(cls, 'swap', None)
        if swap and issubclass(swap, exceptions.APIException):
            mcs.errors_4_swapping[swap] = cls

        return cls


class Error(Exception, metaclass=ErrorMeta):
    """
    base class of errors
    """

    status_code = None
    error_code = None
    error_message = None

    def __init__(self, error_message=None):
        if error_message:
            self.error_message = error_message

    def get_data(self, data=None):
        resp = {
            ERROR_CODE_NAME: self.error_code,
            ERROR_MESSAGE_NAME: self.error_message,
        }
        if isinstance(data, dict):
            resp.update(data)
        elif isinstance(data, (list, tuple)):
            resp.update(results=data)

        return resp

    def get_response(self, data=None, headers=None):
        return response.Response(
            data=self.get_data(data),
            headers=headers,
            status=self.status_code,
        )


class Success(Error):
    """
    this is the special class for http status_code: 200.
    """

    status_code = status.HTTP_200_OK
    error_code = 2000
    error_message = 'success'


class CreatedSuccess(Error):
    status_code = status.HTTP_201_CREATED
    error_code = 2010
    error_message = 'success'


class DestroyedSuccess(Error):
    status_code = status.HTTP_204_NO_CONTENT
    error_code = 2040
    error_message = 'success'


class BadRequestError(Error):
    status_code = status.HTTP_400_BAD_REQUEST
    error_code = 4000
    error_message = 'bad request'
    swap = exceptions.ValidationError


class UnauthorizedError(Error):
    status_code = status.HTTP_401_UNAUTHORIZED
    error_code = 4010
    error_message = 'unauthorized'
    swap = exceptions.NotAuthenticated


class AuthorizeFailedError(Error):
    status_code = status.HTTP_401_UNAUTHORIZED
    error_code = 4011
    error_message = 'authorize failed'
    swap = exceptions.AuthenticationFailed


class PermissionDeniedError(Error):
    status_code = status.HTTP_403_FORBIDDEN
    error_code = 4030
    error_message = 'permission denied'
    swap = exceptions.PermissionDenied


class NotFoundError(Error):
    status_code = status.HTTP_404_NOT_FOUND
    error_code = 4040
    error_message = 'not found'
    swap = exceptions.NotFound


class MethodNotAllowedError(Error):
    status_code = status.HTTP_405_METHOD_NOT_ALLOWED
    error_code = 4050
    error_message = 'method not allowed'
    swap = exceptions.MethodNotAllowed


class NotAcceptableError(Error):
    status_code = status.HTTP_406_NOT_ACCEPTABLE
    error_code = 4060
    error_message = 'not acceptable'
    swap = exceptions.NotAcceptable


class ConflictError(Error):
    status_code = status.HTTP_409_CONFLICT
    error_code = 4090
    error_message = 'data conflict occurred'


class UnsupportedMediaTypeError(Error):
    status_code = status.HTTP_415_UNSUPPORTED_MEDIA_TYPE
    error_code = 4150
    error_message = 'media type not supported'
    swap = exceptions.UnsupportedMediaType


class TooManyRequestsError(Error):
    status_code = status.HTTP_429_TOO_MANY_REQUESTS
    error_code = 4290
    error_message = 'too many requests'
    swap = exceptions.Throttled


class InternalServerError(Error):
    status_code = status.HTTP_500_INTERNAL_SERVER_ERROR
    error_code = 5000
    error_message = 'internal server error'
    swap = exceptions.APIException


class InternalRPCError(Error):
    status_code = status.HTTP_500_INTERNAL_SERVER_ERROR
    error_code = 5001
    error_message = 'internal RPC error'



from django.conf import settings

MB_DRF_EXTENSIONS_CONFIG = getattr(settings, 'MB_DRF_EXTENSIONS', {})
MB_AUTHENTICATIONS = MB_DRF_EXTENSIONS_CONFIG.get('AUTHENTICATIONS', {})
MB_EXCEPTIONS = MB_DRF_EXTENSIONS_CONFIG.get('EXCEPTIONS', {})
MB_VIEWS = MB_DRF_EXTENSIONS_CONFIG.get('VIEWS', {})



import datetime
import urllib.parse

import bson
import requests

from django.core.cache import cache
from django.utils.translation import ugettext_lazy as _
from rest_framework import (
    authentication,
    exceptions,
    status,
)

from .settings import MB_AUTHENTICATIONS


class AnonymousUser(object):
    id = None
    pk = id
    username = ''
    is_staff = False
    is_active = False
    is_superuser = False

    def __init__(self):
        pass

    def __str__(self):
        return 'AnonymousUser'

    def __eq__(self, other):
        return isinstance(other, self.__class__)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return 1  # instances always return the same hash value

    def get_username(self):
        return self.username

    @property
    def is_anonymous(self):
        return True

    @property
    def is_authenticated(self):
        return False


class AuthenticatedUser(object):
    is_staff = False
    is_active = True
    is_superuser = False

    def __init__(self, user, access_token=None):
        self.user = user
        self.access_token = access_token

    def __str__(self):
        return 'AuthenticatedUser'

    def __eq__(self, other):
        return isinstance(other, self.__class__)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return self.pk

    def get_username(self):
        return self.user.get('username')

    @property
    def id(self):
        return self.user.get('uuid')

    @property
    def pk(self):
        return self.user.get('uuid')

    @property
    def is_anonymous(self):
        return False

    @property
    def is_authenticated(self):
        return True

    @property
    def scopes(self):
        return self.user.get('scopes', [])

    @property
    def uuid(self):
        return self.user.get('uuid')

    @property
    def username(self):
        return self.user.get('username')


class OAuth2Mixin(object):
    oauth2_cache_prefix = MB_AUTHENTICATIONS.get(
        'CACHE_PREFIX', 'oauth2')
    oauth2_client_id = MB_AUTHENTICATIONS.get(
        'OAUTH2_CLIENT_ID', '')
    oauth2_client_secret = MB_AUTHENTICATIONS.get(
        'OAUTH2_CLIENT_SECRET', '')
    oauth2_url_scheme = MB_AUTHENTICATIONS.get(
        'OAUTH2_URL_SCHEME', 'http')
    oauth2_url_domain = MB_AUTHENTICATIONS.get(
        'OAUTH2_URL_DOMAIN', 'localhost')
    oauth2_login_path = MB_AUTHENTICATIONS.get(
        'OAUTH2_LOGIN_PATH', 'oauth2/token/')
    oauth2_introspect_path = MB_AUTHENTICATIONS.get(
        'OAUTH2_INTROSPECT_PATH', 'oauth2/introspect/')
    oauth2_login_url = '{}://{}/{}'.format(
        oauth2_url_scheme, oauth2_url_domain, oauth2_login_path)
    oauth2_introspect_url = '{}://{}/{}'.format(
        oauth2_url_scheme, oauth2_url_domain, oauth2_introspect_path)

    def make_introspect_cache_key(self):
        return '{}:{}'.format(
            self.oauth2_cache_prefix,
            self.oauth2_client_id,
        )

    def make_access_cache_key(
            self,
            access_token,
    ):
        return '{}:{}'.format(
            self.oauth2_cache_prefix,
            access_token,
        )

    def save_user_to_cache(
            self,
            access_token,
            user,
            expire,
            mutex=None,
    ):
        if not expire:  # cache user if need be
            return

        delta = (
                datetime.datetime.fromtimestamp(
                    int(expire), datetime.timezone.utc) -
                datetime.datetime.now(datetime.timezone.utc)
        )
        expires_in = int(delta.total_seconds())
        if expires_in < 60:
            return

        expires_in -= 30
        expires_in = min(300, expires_in)
        cache.set(
            self.make_access_cache_key(access_token),
            bson.dumps(user),
            expires_in,
        )

        user_uuid = user.get('uuid')
        if not mutex or not user_uuid:
            return

        access_mutex_cache_key = '{}:{}:{}'.format(
            self.oauth2_cache_prefix, mutex, user_uuid)
        access_token_prev = cache.get(access_mutex_cache_key)
        cache.set(access_mutex_cache_key, access_token)
        if access_token_prev:
            cache.delete(self.make_access_cache_key(access_token_prev))

    def load_user_from_cache(self, access_token):
        data = cache.get(self.make_access_cache_key(access_token))
        return data and bson.loads(data) or {}

    def do_get_introspect_token(self):
        response = requests.post(self.oauth2_login_url, data={
            'grant_type': 'client_credentials',
            'client_id': self.oauth2_client_id,
            'client_secret': self.oauth2_client_secret,
        })
        if response.status_code != status.HTTP_200_OK:
            return None

        response = response.json()
        if response.get('error'):
            return None

        access_token = response.get('access_token')
        expires_in = int(response.get('expires_in'))
        expires_in = expires_in - 30
        if expires_in > 0:
            cache.set(
                self.make_introspect_cache_key(),
                access_token,
                expires_in,
            )

        return access_token

    def get_introspect_token(self):
        return (
                cache.get(self.make_introspect_cache_key()) or
                self.do_get_introspect_token()
        )

    def do_get_user_by_access_token(self, access_token):
        response = requests.post(
            self.oauth2_introspect_url,
            data={'token': access_token},
            headers={
                'Authorization': 'Bearer {}'.format(
                    self.get_introspect_token())
            }
        )
        if not response.ok:
            if response.status_code == status.HTTP_401_UNAUTHORIZED:
                # try refreshing introspect token
                self.do_get_introspect_token()
            return None

        response = response.json()
        scopes = response.get('scope')
        user = response.get('user') or {}
        if response.get('error'):
            return None

        scopes = scopes and scopes.split(' ') or []
        scopes = list(filter(None, scopes))
        user.update(scopes=scopes)

        self.save_user_to_cache(
            access_token, user,
            response.get('exp'),
            mutex=response.get('mutex')
        )
        return user

    def get_user_by_access_token(self, access_token):
        return (
                self.load_user_from_cache(access_token) or
                self.do_get_user_by_access_token(access_token)
        )

    def authorized_request(
            self,
            path,
            method='post',
            params=None,
            data=None,
            access_token=None,
            **kwargs
    ):
        bits = urllib.parse.urlsplit(path)
        location = (
            '{}://{}/{}'.format(
                self.oauth2_url_scheme, self.oauth2_url_domain, path)
            if not (bits.scheme and bits.netloc) else path
        )
        handler = getattr(requests, method.lower())
        response = handler(
            location,
            params=params,
            data=data,
            headers={
                'Authorization': 'Bearer {}'.format(
                    access_token or self.get_introspect_token())
            },
            **kwargs
        )

        try:
            response_json = response.json()
        except ValueError:
            response_json = {'error': response.text}

        return response.status_code, response_json


class BearerAuthentication(
    OAuth2Mixin,
    authentication.BaseAuthentication,
):

    def authenticate(self, request):
        auth = authentication.get_authorization_header(request).split()
        if not auth:
            return None

        if (
                auth[0].lower().decode() !=
                self.authenticate_header(request).lower()
        ):
            return None

        if len(auth) == 1:
            msg = _(
                'Invalid token header. No credentials provided.')
            raise exceptions.AuthenticationFailed(msg)
        elif len(auth) > 2:
            msg = _(
                'Invalid token header. '
                'Token string should not contain spaces.')
            raise exceptions.AuthenticationFailed(msg)

        try:
            token = auth[1].decode()
        except UnicodeError:
            msg = _(
                'Invalid token header. '
                'Token string should not contain invalid characters.')
            raise exceptions.AuthenticationFailed(msg)

        credentials = self.authenticate_credentials(token)
        if not credentials:
            msg = _(
                'Invalid token header. No user found.')
            raise exceptions.AuthenticationFailed(msg)

        return (
            credentials[0],
            credentials[1],
        )

    def authenticate_credentials(self, key):
        user = self.get_user_by_access_token(key)
        return user and (AuthenticatedUser(
            user, access_token=key), key) or None

    def authenticate_header(self, request):
        return 'Bearer'



import json
import uuid

from requests_mock import Mocker

from .authentication import OAuth2Mixin


class AuthenticationMock(Mocker):
    def __init__(
            self,
            username,
            user_uuid=None,
            user_scope=None,
            access_token=None,
    ):
        super().__init__()
        self.username = username
        self.user_uuid = user_uuid or uuid.uuid4().hex
        self.user_scope = (
            ' '.join(user_scope) if
            isinstance(user_scope, (list, tuple)) else ''
        )
        self.access_token = access_token or uuid.uuid4().hex

    def start(self):
        super().start()
        self.register_uri(
            'POST', OAuth2Mixin.oauth2_login_url,
            text=json.dumps({
                'access_token': self.access_token,
                'expires_in': 36000
            }),
        )
        self.register_uri(
            'POST', OAuth2Mixin.oauth2_introspect_url,
            text=json.dumps({
                'scope': self.user_scope,
                'user': {
                    'username': self.username,
                    'uuid': self.user_uuid,
                }
            }),
        )



from uuid import UUID

from django.forms import CharField
from django.core.exceptions import ValidationError
from django.utils.translation import gettext_lazy as _
from django_filters.rest_framework import filters


class UUIDListField(CharField):
    default_error_messages = {
        'invalid': _('Enter a valid list of UUID.'),
    }

    def __init__(self, separator=',', *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.separator = separator

    def prepare_value(self, value):
        if isinstance(value, list):
            return self.separator.join(
                v.hex for
                v in value if isinstance(v, UUID)
            )
        return value

    def to_python(self, value):
        value = super().to_python(value)
        if value in self.empty_values:
            return None

        if not isinstance(value, list):
            try:
                value = [
                    UUID(v.strip()) for
                    v in value.split(self.separator) if
                    v.strip()
                ]
            except ValueError:
                raise ValidationError(
                    self.error_messages['invalid'],
                    code='invalid'
                )
        return value


class CharListField(CharField):
    default_error_messages = {
        'invalid': _('Enter a valid list of Text.'),
    }

    def __init__(self, separator=',', *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.separator = separator

    def prepare_value(self, value):
        if isinstance(value, list):
            return self.separator.join([
                v for v in value if
                v.strip() and self.separator not in v
            ])
        return value

    def to_python(self, value):
        value = super().to_python(value)
        if value in self.empty_values:
            return None

        if not isinstance(value, list):
            try:
                value = [
                    v.strip() for
                    v in value.split(self.separator) if
                    v.strip()
                ]
            except ValueError:
                raise ValidationError(
                    self.error_messages['invalid'],
                    code='invalid'
                )
        return value


class UUIDListFilter(filters.Filter):
    field_class = UUIDListField


class CharListFilter(filters.Filter):
    field_class = CharListField



from rest_framework import (
    pagination,
    settings,
)
from .settings import MB_VIEWS

PAGE_SIZE = MB_VIEWS.get(
    'PAGE_SIZE', settings.api_settings.PAGE_SIZE)
PAGE_SIZE_MAX = MB_VIEWS.get(
    'PAGE_SIZE_MAX', 1000)


class CursorPagination(pagination.CursorPagination):
    page_size = PAGE_SIZE
    max_page_size = PAGE_SIZE_MAX


class LimitOffsetPagination(pagination.LimitOffsetPagination):
    default_limit = PAGE_SIZE
    max_limit = PAGE_SIZE_MAX


class PageNumberPagination(pagination.PageNumberPagination):
    page_size = PAGE_SIZE
    max_page_size = PAGE_SIZE_MAX



from django.db import transaction
from rest_framework import mixins

from .exceptions import Success
from .settings import (
    MB_EXCEPTIONS,
    MB_VIEWS,
)


class AtomicMixin(object):
    atomic_actions = MB_VIEWS.get(
        'ATOMIC_ACTIONS', [
            'create',
            'update',
            'destroy',
            'post',
            'put',
            'patch',
            'delete',
        ]
    )
    atomic_enabled = bool(
        MB_VIEWS.get('ATOMIC_ENABLED', False))
    atomic_savepoint = bool(
        MB_VIEWS.get('ATOMIC_SAVEPOINT', True))

    def __init__(self, *args, **kwargs):
        self.patch_actions()

    def patch_actions(self):
        if not self.atomic_enabled:
            return  # do nothing if atomic is not enabled

        if not self.atomic_actions:
            dispatch_method = getattr(self, 'dispatch', None)
            if callable(dispatch_method):
                setattr(
                    self,
                    'dispatch',
                    transaction.atomic(
                        dispatch_method,
                        savepoint=self.atomic_savepoint,
                    )
                )

        else:
            for action in self.atomic_actions:
                action_method = getattr(self, action, None)
                if callable(action_method):
                    setattr(
                        self,
                        action,
                        transaction.atomic(
                            action_method,
                            savepoint=self.atomic_savepoint,
                        ),
                    )

    @classmethod
    def as_view(cls, *args, **kwargs):
        as_view_method = getattr(super(), 'as_view', None)
        if not callable(as_view_method):
            raise AssertionError('no as_view method found')

        view = as_view_method(*args, **kwargs)
        return (
            transaction.non_atomic_requests(view) if
            cls.atomic_enabled else view
        )


class ResponseMixin(object):
    error_return = Success()
    error_return_ensured = bool(
        MB_EXCEPTIONS.get('ERROR_RETURN_ENSURED', False))

    def get_response(self, resp):
        if self.error_return_ensured:
            resp.data = self.error_return.get_data(resp.data)
        return resp


class CreateModelMixin(
    ResponseMixin,
    mixins.CreateModelMixin,
):

    def create(self, request, *args, **kwargs):
        return self.get_response(
            super().create(request, *args, **kwargs))


class DestroyModelMixin(
    ResponseMixin,
    mixins.DestroyModelMixin,
):

    def destroy(self, request, *args, **kwargs):
        return self.get_response(
            super().destroy(request, *args, **kwargs))


class UpdateModelMixin(
    ResponseMixin,
    mixins.UpdateModelMixin,
):

    def update(self, request, *args, **kwargs):
        return self.get_response(
            super().update(request, *args, **kwargs))


class ListModelMixin(
    ResponseMixin,
    mixins.ListModelMixin,
):

    def list(self, request, *args, **kwargs):
        return self.get_response(
            super().list(request, *args, **kwargs))


class RetrieveModelMixin(
    ResponseMixin,
    mixins.RetrieveModelMixin,
):

    def retrieve(self, request, *args, **kwargs):
        return self.get_response(
            super().retrieve(request, *args, **kwargs))



#!/usr/bin/env python
import re

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

with open('mb_drf_extensions/__init__.py', 'r') as fd:
    version = re.search(
        r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]',
        fd.read(), re.MULTILINE).group(1)

if not version:
    raise RuntimeError('Cannot find version information')

with open('README.md', 'rb') as f:
    readme = f.read().decode('utf-8')

setup(
    name='mb-drf-extensions',
    version=version,
    description='Mind Bricks Django Rest Framework Extensions',
    long_description=readme,
    packages=[
        'mb_drf_extensions',
    ],
    install_requires=[
        'bson>=0.5.0',
        'django>=1.11.0',
        'django-filter>=2.4.0',
        'djangorestframework>=3.7.0',
        'requests>=2.0.0',
        'requests-mock>=1.8.0',
    ],
    include_package_data=True,
    url='https://www.mind-bricks.com',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Framework :: Django',
    ],
)



from rest_framework import permissions


class ScopesPermission(permissions.BasePermission):
    def has_permission(self, request, view):
        required_scopes = self.get_scopes(request, view)
        if not required_scopes:
            # no required scopes found
            return True

        scopes = getattr(request.user, 'scopes', [])
        if not scopes:
            return False

        # any matched scope are allowed
        for required_scope in required_scopes:
            if frozenset(required_scope.split()).issubset(scopes):
                return True

        return False

    def get_scopes(self, request, view):
        try:
            return getattr(view, 'required_scopes')
        except AttributeError:
            raise AssertionError(
                'TokenHasScope requires the view to define '
                'the required_scopes attribute'
            )



from django.core.exceptions import PermissionDenied
from django.http import Http404
from rest_framework import (
    exceptions,
    views,
)
from .exceptions import (
    ErrorMeta,
    Error,
    InternalServerError,
)
from .mixins import AtomicMixin


def exception_handler(exc, _context):
    """
    Returns the response that should be used for any given exception.

    By default we handle the REST framework `APIException`, and also
    Django's built-in `Http404` and `PermissionDenied` exceptions.

    Any unhandled exceptions may return `None`, which will cause a 500 error
    to be raised.
    """
    if isinstance(exc, Http404):
        exc = exceptions.NotFound()
    elif isinstance(exc, PermissionDenied):
        exc = exceptions.PermissionDenied()

    if isinstance(exc, exceptions.APIException):
        # swap exception
        exc_cls = ErrorMeta.errors_4_swapping.get(exc.__class__)
        exc_cls = exc_cls or InternalServerError
        exc = exc_cls(exc.detail)

    if isinstance(exc, Error):
        headers = {}

        exec_auth_header = getattr(exc, 'auth_header', None)
        if exec_auth_header:
            headers['WWW-Authenticate'] = exec_auth_header

        exec_wait = getattr(exc, 'wait', None)
        if exec_wait:
            headers['Retry-After'] = '%d' % exec_wait

        views.set_rollback()

        return exc.get_response(headers=headers)

    return None


class APIView(
    AtomicMixin,
    views.APIView,
):
    pass



from rest_framework import viewsets

from .mixins import (
    AtomicMixin,
    CreateModelMixin,
    DestroyModelMixin,
    ListModelMixin,
    RetrieveModelMixin,
    UpdateModelMixin,
)


class GenericViewSet(
    AtomicMixin,
    viewsets.GenericViewSet,
):
    pass


class ReadOnlyModelViewSet(
    ListModelMixin,
    RetrieveModelMixin,
    GenericViewSet,
):
    pass


class ModelViewSet(
    CreateModelMixin,
    RetrieveModelMixin,
    UpdateModelMixin,
    DestroyModelMixin,
    ListModelMixin,
    GenericViewSet,
):
    pass
